@@ -1,12 +1,21 @@
 import hashlib
 import logging
 import time
-from .connection import StudyClient
+from typing import Optional, Union
+from .connection import StudyClient, Experiment, Subject, State, Workflow, Action, Transition, ExternalSystem
+from . import exceptions
 
-__version__ = '0.2.0'
+__version__ = '0.4.0'
+__all__ = ['StudyClient', 'Experiment', 'Subject', 'State', 'Workflow', 'Action', 'Transition', 'ExternalSystem', 'connect']
 
 
-def connect(uri, user=None, password=None, logger=None, loglevel=None, debug=False, user_agent_prefix=None):
+def connect(uri: str,
+            user: Optional[str]=None,
+            password: Optional[str]=None,
+            logger: Optional[logging.Logger]=None,
+            loglevel: Optional[Union[str, int]]=None,
+            debug: bool=False,
+            user_agent_prefix: Optional[str]=None) -> StudyClient:
     # Generate a hash for the connection
     hasher = hashlib.md5()
     hasher.update(uri.encode('utf-8'))
