@@ -5,17 +5,23 @@ import os
 import netrc
 import platform
 import requests
+from urllib import parse
 
-import six
-from six.moves.urllib import parse
+from typing import List, Mapping, Optional, Tuple, Type, TypeVar, Union, Dict
 
-from .core import Experiment, Subject, State, Workflow, Action, Transition
+from .core import StudyClientBaseObject, Experiment, Subject, State, Workflow, Action, Transition, ExternalSystem
 from .drivers import DRIVERS
 from .exceptions import StudyGovernorValueError
 from .exceptions import StudyGovernorNoAuthError
 from .exceptions import StudyGovernorResponseError
+from .exceptions import StudyGovernorNotFoundError
 from .exceptions import StudyGovernorSSLError
 from .exceptions import StudyGovernorConnectionError
+from .exceptions import SubjectExistsError
+from .exceptions import ExperimentExistsError
+
+
+JSONType = Union[str, int, float, bool, None, Mapping[str, 'JSON'], List['JSON']]
 
 
 class StudyClient(object):
@@ -26,6 +32,7 @@ class StudyClient(object):
         Workflow: '__workflows__',
         Action: '__actions__',
         Transition: '__transitions__',
+        ExternalSystem: '__external_systems__'
     }
 
     def __init__(self, server, user=None, password=None, session=None, logger=None,
@@ -102,13 +109,16 @@ class StudyClient(object):
         # Detect API version and load appropriate driver
         self.detect_api_version()
 
+        # Check whether login was successful
+        self._test_login()
+
         try:
-            response = self.get('/tags', accepted_status=[200, 401, 404], timeout=2)
+            response = self.get('/workflows', accepted_status=[200, 401], timeout=2)
         except Exception as exception:
-            self.logger.error(f'Error in test tags request {self.original_uri}: {exception}')
+            self.logger.error(f'Error in test workflows request {self.original_uri}: {exception}')
             raise
 
-        self.logger.info(f'Response code on test tags request: {response.status_code}')
+        self.logger.info(f'Response code on test workflows request: {response.status_code}')
         if response.status_code == 401:
             payload = response.headers['www-authenticate'].split("=")[1].strip("\"")
             raise StudyGovernorNoAuthError(payload, f'Could not authenticate for {self.original_uri}')
@@ -128,6 +138,25 @@ class StudyClient(object):
         """
         pass
 
+    def _test_login(self):
+        try:
+            response = self.get('/subjects', accepted_status=[200, 401, 404], timeout=2)
+        except Exception as exception:
+            message = f'Cannot create connection to {self.original_uri}'
+            if self.debug:
+                message = '{}: {}'.format(message, exception)
+            self.logger.error(message)
+            raise StudyGovernorConnectionError(message)
+
+        self.logger.info(f'Response code on initial try: {response.status_code}')
+        if response.status_code == 401:
+            payload = response.headers['www-authenticate'].split("=")[1].strip("\"")
+            raise StudyGovernorNoAuthError(payload, f'Could not authenticate for {self.original_uri}')
+
+        if b'<div class="login-form">' in response.content:
+            payload = response.headers['www-authenticate'].split("=")[1].strip("\"")
+            raise StudyGovernorNoAuthError(payload, f'Could not authenticate for {self.original_uri}')
+
     def detect_api_version(self):
         # Detect API version
         try:
@@ -185,7 +214,9 @@ class StudyClient(object):
         # Select API Driver
         self.driver = DRIVERS[self.api_version](self)
 
-    def _format_uri(self, path, query=None):
+    def _format_uri(self,
+                    path: str,
+                    query: Optional[Mapping[str, str]]=None) -> str:
         if path[0] != '/':
 
             if self._url_prefix is not None and path.startswith(self._url_prefix):
@@ -204,7 +235,7 @@ class StudyClient(object):
             query = {}
 
         # Sanitize unicode in query
-        query = {k: v.encode('utf-8', 'xmlcharrefreplace') if isinstance(v, six.text_type) else v for k, v in query.items()}
+        query = {k: v.encode('utf-8', 'xmlcharrefreplace') if isinstance(v, str) else v for k, v in query.items()}
 
         # Create the query string
         if len(query) > 0:
@@ -219,9 +250,30 @@ class StudyClient(object):
                 query_string,
                 '')
 
-        return parse.urlunparse(data)
+        result = parse.urlunparse(data)
 
-    def _check_response(self, response, accepted_status=None, uri=None):
+        # Make sure we work with strings
+        if isinstance(result,  bytes):
+            result = result.decode('utf-8')
+
+        return result
+
+    def url_for(self,
+                obj: StudyClientBaseObject,
+                query: Optional[Mapping[str, str]]=None,
+                scheme: Optional[str]=None) -> str:
+        """
+        Return the (external) url for a given XNAT object
+        :param XNATBaseObject obj: object to get url for
+        :param query: extra query string parameters
+        :return: external url for the object
+        """
+        return self._format_uri(obj.uri, query=query)
+
+    def _check_response(self,
+                        response: requests.Response,
+                        accepted_status: Optional[List[int]]=None,
+                        uri: Optional[str]=None):
 
         if self.debug:
             self.logger.debug(f'Received response with status code: {response.status_code}')
@@ -233,7 +285,12 @@ class StudyClient(object):
             if response.status_code not in accepted_status:
                 raise StudyGovernorResponseError(uri, response.status_code, accepted_status)
 
-    def get(self, path, query=None, accepted_status=None, timeout=None, headers=None):
+    def get(self,
+            path: str,
+            query: Optional[Mapping[str, str]]=None,
+            accepted_status: Optional[List[int]]=None,
+            timeout: Union[float, Tuple[float, float]]=None,
+            headers: Optional[Mapping[str, str]]=None) -> requests.Response:
         """
         Retrieve the content of a given REST directory.
 
@@ -262,23 +319,28 @@ class StudyClient(object):
         self._check_response(response, accepted_status=accepted_status, uri=uri)  # Allow OK, as we want to get data
         return response
 
-    def get_json(self, uri, query=None, accepted_status=None):
+    def get_json(self,
+                 path: str,
+                 query: Optional[Mapping[str, str]]=None,
+                 accepted_status: Optional[List[int]]=None) -> JSONType:
         """
         Helper function that perform a GET, but sets the format to JSON and
         parses the result as JSON
 
-        :param str uri: the path of the uri to retrieve (e.g. "/data/archive/projects")
+        :param str path: the path of the uri to retrieve (e.g. "/data/archive/projects")
                          the remained for the uri is constructed automatically
         :param list accepted_status: a list of the valid values for the return code, default [200]
         :param dict query: the values to be added to the query string in the uri
         """
-        response = self.get(uri, query=query, accepted_status=accepted_status)
+        response = self.get(path, query=query, accepted_status=accepted_status)
         return response.json()
 
-    def change_lock(self, lock_uri, value):
-        return self.driver.change_lock(lock_uri, value)
-
-    def head(self, path, accepted_status=None, allow_redirects=False, timeout=None, headers=None):
+    def head(self,
+             path: str,
+             accepted_status: Optional[List[int]]=None,
+             allow_redirects: bool=False,
+             timeout: Union[float, Tuple[float, float]]=None,
+             headers: Optional[Mapping[str, str]]=None):
         """
         Retrieve the header for a http request of a given REST directory.
 
@@ -307,7 +369,14 @@ class StudyClient(object):
         self._check_response(response, accepted_status=accepted_status, uri=uri)  # Allow OK, as we want to get data
         return response
 
-    def post(self, path, data=None, json=None, query=None, accepted_status=None, timeout=None, headers=None):
+    def post(self,
+             path: str,
+             data=None,
+             json=None,
+             query: Optional[Mapping[str, str]]=None,
+             accepted_status: Optional[List[int]]=None,
+             timeout: Union[float, Tuple[float, float]]=None,
+             headers: Optional[Mapping[str, str]]=None):
         """
         Post data to a given REST directory.
 
@@ -340,7 +409,15 @@ class StudyClient(object):
         self._check_response(response, accepted_status=accepted_status, uri=uri)
         return response
 
-    def put(self, path, data=None, files=None, json=None, query=None, accepted_status=None, timeout=None, headers=None):
+    def put(self,
+            path: str,
+            data=None,
+            files=None,
+            json=Mapping,
+            query: Optional[Mapping[str, str]]=None,
+            accepted_status: Optional[List[int]]=None,
+            timeout: Union[float, Tuple[float, float]]=None,
+            headers: Optional[Mapping[str, str]]=None):
         """
         Put the content of a given REST directory.
 
@@ -381,7 +458,12 @@ class StudyClient(object):
         self._check_response(response, accepted_status=accepted_status, uri=uri)  # Allow created OK or Create status (OK if already exists)
         return response
 
-    def delete(self, path, headers=None, accepted_status=None, query=None, timeout=None):
+    def delete(self,
+               path: str,
+               headers: Optional[Mapping[str, str]]=None,
+               accepted_status: Optional[List[int]]=None,
+               query: Optional[Mapping[str, str]]=None,
+               timeout: Union[float, Tuple[float, float]]=None):
         """
         Delete the content of a given REST directory.
 
@@ -419,12 +501,16 @@ class StudyClient(object):
         self._cache.clear()
         self._cache = {x: {} for x in self.CACHE_NAMES.values()}
 
-    def create_object(self, cls, data):
+    # Define a type variable to get the create-object hinting correctly
+    T = TypeVar('T')
+
+    def create_object(self, cls: Type[T], data: Union[Mapping, str]) -> T:
         if isinstance(data, str):
             uri = data
             data = {}
         else:
             # TODO: hack due to inconsistency in study-governor, needs to be fixed
+            data = dict(data)
             if 'uri' not in data and 'url' in data:
                 data['uri'] = data.pop('url')
 
@@ -439,6 +525,96 @@ class StudyClient(object):
             self._cache[type_name][uri] = task
             return task
 
-    def get_subjects(self):
+    def get_subject(self, label: str) -> Subject:
+        try:
+            data = self.get_json(f'/subjects/{label}')
+        except StudyGovernorResponseError as exception:
+            if exception.status_code == 404:
+                raise StudyGovernorNotFoundError(
+                    Subject,
+                    exception.uri,
+                    exception.status_code,
+                    exception.acceptable_status_codes
+                )
+            else:
+                raise
+
+        return self.create_object(Subject, data)
+
+    def get_subjects(self) -> List[Subject]:
         data = self.get_json('/subjects')
         return [self.create_object(Subject, x) for x in data['subjects']]
+
+    def create_subject(self,
+                       subject_label: str,
+                       date_of_birth: Optional[str]=None,
+                       external_ids: Optional[Mapping[str, str]]=None,
+                       exist_ok: bool=False):
+        try:
+            # Check if subject exists and handle accordingly
+            subject = self.get_subject(subject_label)
+
+            if not exist_ok:
+                raise SubjectExistsError(f'Cannot create subject {subject_label}, it already exists!')
+
+            if date_of_birth:
+                subject.date_of_birth = date_of_birth
+        except StudyGovernorNotFoundError:
+            # Could not find subject, safely create it
+            subject = Subject(self, label=subject_label, date_of_birth=date_of_birth)
+
+        if external_ids:
+            for key, value in external_ids.items():
+                subject.external_ids[key] = value
+
+        return subject
+
+    def get_experiment(self, label: str) -> Experiment:
+        try:
+            data = self.get_json(f'/experiments/{label}')
+        except StudyGovernorResponseError as exception:
+            if exception.status_code == 404:
+                raise StudyGovernorNotFoundError(
+                    Experiment,
+                    exception.uri,
+                    exception.status_code,
+                    exception.acceptable_status_codes
+                )
+            else:
+                raise
+
+        return self.create_object(Experiment, data)
+
+    def create_experiment(self,
+                          subject: Union[Subject, str],
+                          label: str,
+                          scandate: Optional[str]=None,
+                          workflow: Optional[Union[Workflow, str]]=None,
+                          external_ids: Optional[Mapping[str, str]]=None,
+                          exist_ok: bool=False):
+        try:
+            # Check if experiment exists and handle accordingly
+            experiment = self.get_experiment(label)
+
+            if not exist_ok:
+                raise ExperimentExistsError(f'Cannot create experiment {label}, it already exists!')
+
+        except StudyGovernorNotFoundError:
+            # Could not find experiment, safely create it
+            experiment = Experiment(self,
+                                    label=label,
+                                    subject=subject,
+                                    scandate=scandate,
+                                    workflow=workflow)
+
+        if external_ids:
+            for key, value in external_ids.items():
+                experiment.external_ids[key] = value
+
+        return experiment
+
+    @property
+    def external_systems(self) -> Dict[str, ExternalSystem]:
+        data = self.get_json('/external_systems')
+        data = [self.create_object(ExternalSystem, x) for x in data['external_systems']]
+        return {x.system_name: x for x in data}
