@@ -20,6 +20,15 @@ class StudyGovernorResponseError(StudyGovernorError):
         )
 
 
+class StudyGovernorNotFoundError(StudyGovernorResponseError):
+    def __init__(self, cls, uri, status_code, acceptable_status_codes):
+        super().__init__(uri, status_code, acceptable_status_codes)
+        self.cls = cls
+
+    def __str__(self):
+        return f'Could not find {self.cls.__name__} using {self.uri}, encountered {self.status_code} status code!'
+
+
 class StudyGovernorConnectionError(StudyGovernorError):
     pass
 
@@ -34,3 +43,10 @@ class StudyGovernorSSLError(StudyGovernorConnectionError):
     pass
 
 
+class SubjectExistsError(StudyGovernorValueError):
+    pass
+
+
+class ExperimentExistsError(StudyGovernorValueError):
+    pass
+
